/*
 * Copyright (C) 2015 PHYTEC Messtechnik GmbH,
 * Author:   Alexander Fratzer <a.fratzer@phytec.de>
 *           Alexander Aring   <a.aring@phytec.de>
 * Released under the MIT license
 */

#include "compare_raw_image.h"

/*-----------------------------------------------------------------------------
 *  image_file_path for parameter n with defaultpath
 *-----------------------------------------------------------------------------*/
static char image_file_path[PATH_MAX] = DEFAULT_IMAGE_FILE;
/*-----------------------------------------------------------------------------
 *  image_ref_file_path for parameter m with defaultpath.
 *  this is for the reference image.
 *-----------------------------------------------------------------------------*/
static char image_ref_file_path[PATH_MAX] = DEFAULT_IMAGE_REF_FILE;

/*-----------------------------------------------------------------------------
 *  compare bytewise a image.
 *-----------------------------------------------------------------------------*/
static __u64
compare_raw_images(const void *buffer, const void *buffer_ref, size_t size)
{
	__u64 ret = 0;
	int i;

	for (i = 0; i < size; i++) {
		ret += abs(((__u8 *) buffer)[i] - ((__u8 *) buffer_ref)[i]);
	}

	return ret;
}

static int get_file_size(char *path, off_t * filesize)
{
	int err;
	struct stat statbuf;

	err = stat(path, &statbuf);
	if (err == -1) {
		perror("stat");
		return -1;
	}

	*filesize = statbuf.st_size;

	return 0;
}

/*-----------------------------------------------------------------------------
 *  read file into buf, caller need to call free on success.
 *-----------------------------------------------------------------------------*/
static int file_read_into_buffer(char *path, void *buf, size_t size)
{
	int err;
	size_t count;
	FILE *file = NULL;

	file = fopen(path, "r");
	if (!file) {
		perror("fopen");
		goto out;
	}

	count = fread(buf, size, 1, file);
	if (!count)
		goto close;

	err = fclose(file);
	if (err == EOF) {
		perror("fclose");
		goto out;
	}

	return 0;
 close:
	fclose(file);
 out:
	return -1;
}

int main(int argc, char *argv[])
{
	int opt, err;
	void *buf = NULL, *buf_ref = NULL, *buf_888 = NULL, *buf_ref_888 = NULL;
	off_t filesize, filesize_ref;
	__u64 sum;
	__u32 pixel_format = V4L2_PIX_FMT_GREY;

	/* Getting parameters */
	while ((opt = getopt(argc, argv, "n:m:f:")) != -1) {
		switch (opt) {
		case 'n':
			strncpy(image_file_path, optarg, PATH_MAX);
			break;
		case 'm':
			strncpy(image_ref_file_path, optarg, PATH_MAX);
			break;
		case 'f':
			if (!strcmp("bayer", optarg))
				pixel_format = V4L2_PIX_FMT_SBGGR8;
			if (!strcmp("yuyv", optarg))
				pixel_format = V4L2_PIX_FMT_YUYV;
			if (!strcmp("yvyu", optarg))
				pixel_format = V4L2_PIX_FMT_YVYU;
			if (!strcmp("uyvy", optarg))
				pixel_format = V4L2_PIX_FMT_UYVY;
			if (!strcmp("vyuy", optarg))
				pixel_format = V4L2_PIX_FMT_VYUY;
			if (!strcmp("rgb", optarg))
				pixel_format = V4L2_PIX_FMT_RGB565;
			break;
		default:	/* '?' */
			fprintf(stderr, "Usage: %s " COMPARE_RAW_IMAGE_USAGE,
				argv[0]);
			goto out;
		}
	}

	err = get_file_size(image_file_path, &filesize);
	if (err < 0)
		goto out;

	err = get_file_size(image_ref_file_path, &filesize_ref);
	if (err < 0)
		goto out;

	if (filesize_ref != filesize) {
		fprintf(stderr, "Can't compare, filesizes are not the same!\n");
		goto out;
	}

	buf = malloc(filesize);
	if (err < 0)
		goto out;

	buf_ref = malloc(filesize_ref);
	if (err < 0)
		goto out_free_buf;

	err = file_read_into_buffer(image_file_path, buf, filesize);
	if (err < 0)
		goto out_free;

	err = file_read_into_buffer(image_ref_file_path, buf_ref, filesize_ref);
	if (err < 0)
		goto out_free;

	if (pixel_format == V4L2_PIX_FMT_RGB565) {
		filesize *= 2;
		filesize_ref *= 2;

		buf_888 = malloc(filesize);
		if (err < 0)
			goto out_free;

		buf_ref_888 = malloc(filesize_ref);
		if (err < 0)
			goto out_free_buf888;

		err = rgb565_to_rgb888(buf, filesize / 2, buf_888);
		if (err < 0)
			goto out_free_888;

		err = rgb565_to_rgb888(buf_ref, filesize_ref / 2, buf_ref_888);
		if (err < 0)
			goto out_free_888;

		free(buf);
		free(buf_ref);

		buf = buf_888;
		buf_ref = buf_ref_888;
	}

	sum = compare_raw_images(buf, buf_ref, filesize);

	printf("%llu\n", sum);

	free(buf);
	free(buf_ref);

	return EXIT_SUCCESS;

 out_free_888:
	free(buf_ref_888);
 out_free_buf888:
	free(buf_888);
 out_free:
	free(buf_ref);
 out_free_buf:
	free(buf);
 out:
	return EXIT_FAILURE;
}
