/*
 * Copyright (C) 2015 PHYTEC Messtechnik GmbH,
 * Author:   Alexander Fratzer <a.fratzer@phytec.de>
 *           Alexander Aring   <a.aring@phytec.de>
 * Released under the MIT license
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include <stdint.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include <inttypes.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <linux/videodev2.h>
#include <linux/v4l2-subdev.h>

#include "save_raw_image.h"
#include "videodev.h"
#include "regparse.h"
#include "errhandl.h"
#include "device_utils.h"

#define DROP_FR_MIN 0
#define DROP_FR_MAX 1000

#define SAVE_FR_MIN 1
#define SAVE_FR_MAX 100

#define IN_RANGE(v,l,r) ((v)<(l))||((v)>(r))?false:true
#define IS_BOOL(b) (0 == b || 1 == b)

/**
 * @brief charge number of digits in assign value
 *
 * @param value - value
 *
 * @return - number of digits of value 
 */
static inline uint32_t sri_int_leng(uint32_t value);

/**
 * @brief search pixel-format in global array identified by
 * fmt_name and fills fmp with the associate values.
 *
 * This function does not allocate memory. fmt has to be pre-
 * allocated.
 *
 * @param fmt_name - identification string of pixel-format
 * @param fmt - resulting structure filled with associated
 *			  values.
 *
 * @return - success: index of pm_format in global array
 *		   fail: EINVAL -> unknown string
 */
static int sri_str_to_fmt(char *fmt_name, struct pm_format *fmt);

/**
 * @brief search video-standard in global array identified by
 * std_name and fills std with the associate values.
 *
 * This function does not allocate memory. std has to be pre-
 * allocated.
 *
 * @param std_name - identification string of video-standart
 * @param std - resulting structure filled with associated
 *			  values.
 *
 * @return - success: index of vid_std in global array
 *		   fail: EINVAL -> unknown string
 */
static int sri_str_to_std(char *std_name, struct vid_std *std);

/**
 * @brief print commandline arguments to stream
 *
 * @param s - sri_args structure containing default values
 * @param fp - stream (e.g. stdout)
 * @param argc - number of cmd-line arguements (from main)
 * @param argv - cmd-line string (from main)
 */
static void sri_print_help(struct sri_args *s, FILE * fp, int argc,
			   char **argv);

/**
 * @brief hard-write function to cam register
 *
 * @param devname - device name
 * @param address - address of cam register
 * @param reg - cam register
 * @param val - calue for cam register
 * @param mode - a -> keep value but not set
 *			   o -> set value
 *			   x -> reverse setting of value
 * @param use_subdev - flag to indicate subdevice usage for reg access
 * @param reg_state - not NULL -> saves current reg stare
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int sri_cam_reg_write(uint32_t address, uint64_t reg,
				 uint64_t val, char mode, uint32_t use_subdev);

/**
 * @brief identify name of next image file
 *
 * This function allocates memory. It is in
 * responsibility of the user to free this 
 * memory.
 *
 * @param save_name - name format
 * @param numb - number of fileid
 *
 * @return - sucess: pointer to filname
 *		   fail: ERRORCODE
 */
static char *sri_next_filestr(char *save_name, uint32_t numb);

/**
 * @brief Writes all data from the image array to files
 * using the format. Each image to one file.
 *
 * @param save_name - filename format
 * @param img - array of images 
 * @param img_count - number of elements in img
 *
 * @return - success: 0
 *		   fail: ERRORCODE (cast to pointer)
 */
static int sri_write_to_file(char *save_name, struct image *img,
				 uint32_t img_count);

/**
 * @brief Checks all necessary ranges of setted in sargs.
 *
 * @param sargs - initialized arguement struct
 * 
 * @return - success: 0
 *		   fail: ERRORCODE (ERANGE)
 */
static int sri_range_check(struct sri_args *sargs);

/**
 * @brief Parse cmd-line arguements for necessary operators.
 * Not setted operators will be initialized with default values
 * defined in header files.
 *
 * This function allocates memory. It's in the responsebility
 * of the user to free this memory.
 *
 * @param argc - number of cmd-line arguments (from main)
 * @param argv - cmd-line argument string (from main)
 *
 * @return - success: pointer of filled sri_args structure.
 *		   fail: ERRORCODE (cast to pointer)
 */
static struct sri_args *sri_args_get(int argc, char **argv);

/**
 * @brief Print necessary operators, witch are necessary for the
 * functionalty.
 *
 * @param sargs - pointer of initialized operators
 */
static void sri_print_args(struct sri_args *sargs);

/**
 * @brief Collection of subdevice-initializaions necessary
 * for capturing. 
 *
 * @param sargs - pointer to initialized operators
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int sri_subdevs_init(struct sri_args *sargs);

/**
 * @brief Checks accessibility of camera driver
 */
static int sri_camera_check();

static const struct pm_format pm_formats[] = {
	{"Y8", V4L2_PIX_FMT_GREY, V4L2_MBUS_FMT_Y8_1X8},
	{"Y10", V4L2_PIX_FMT_Y10, V4L2_MBUS_FMT_Y10_1X10},
	{"Y12", V4L2_PIX_FMT_Y12, V4L2_MBUS_FMT_Y12_1X12},
	{"SGRBG8", V4L2_PIX_FMT_SGRBG8, V4L2_MBUS_FMT_SGRBG8_1X8},
	{"SGRBG10", V4L2_PIX_FMT_SGRBG10, V4L2_MBUS_FMT_SGRBG10_1X10},
	{"SGRBG12", V4L2_PIX_FMT_SGRBG12, V4L2_MBUS_FMT_SGRBG12_1X12},
	{"SBGGR8", V4L2_PIX_FMT_SBGGR8, V4L2_MBUS_FMT_SBGGR8_1X8},
	{"SBGGR10", V4L2_PIX_FMT_SBGGR10, V4L2_MBUS_FMT_SBGGR10_1X10},
	{"SBGGR12", V4L2_PIX_FMT_SBGGR12, V4L2_MBUS_FMT_SBGGR12_1X12},
	{"SGBRG8", V4L2_PIX_FMT_SGBRG8, V4L2_MBUS_FMT_SGBRG8_1X8},
	{"SGBRG10", V4L2_PIX_FMT_SGBRG10, V4L2_MBUS_FMT_SGBRG10_1X10},
	{"SGBRG12", V4L2_PIX_FMT_SGBRG12, V4L2_MBUS_FMT_SGBRG12_1X12},
	{"SRGGB8", V4L2_PIX_FMT_SRGGB8, V4L2_MBUS_FMT_SRGGB8_1X8},
	{"SRGGB10", V4L2_PIX_FMT_SRGGB10, V4L2_MBUS_FMT_SRGGB10_1X10},
	{"SRGGB12", V4L2_PIX_FMT_SRGGB12, V4L2_MBUS_FMT_SRGGB12_1X12},
	{"UYVY8", V4L2_PIX_FMT_UYVY, V4L2_MBUS_FMT_UYVY8_2X8},
	{"UYVY10", V4L2_PIX_FMT_UYVY, V4L2_MBUS_FMT_UYVY10_2X10},
	{"UYVY12", V4L2_PIX_FMT_UYVY, V4L2_MBUS_FMT_UYVY12_2X12},
	{"VYUY8", V4L2_PIX_FMT_VYUY, V4L2_MBUS_FMT_VYUY8_2X8},
	{"VYUY10", V4L2_PIX_FMT_VYUY, V4L2_MBUS_FMT_VYUY10_2X10},
	{"VYUY12", V4L2_PIX_FMT_VYUY, V4L2_MBUS_FMT_VYUY12_2X12},
	{"YUYV8", V4L2_PIX_FMT_YUYV, V4L2_MBUS_FMT_YUYV8_2X8},
	{"YUYV10", V4L2_PIX_FMT_YUYV, V4L2_MBUS_FMT_YUYV10_2X10},
	{"YUYV12", V4L2_PIX_FMT_YUYV, V4L2_MBUS_FMT_YUYV12_2X12},
	{"YVYU8", V4L2_PIX_FMT_YVYU, V4L2_MBUS_FMT_YVYU8_2X8},
	{"YVYU10", V4L2_PIX_FMT_YVYU, V4L2_MBUS_FMT_YVYU10_2X10},
	{"YVYU12", V4L2_PIX_FMT_YVYU, V4L2_MBUS_FMT_YVYU12_2X12},
	{"Y16_2X8", V4L2_PIX_FMT_Y16, 0x2041},
	{"\0", 0, 0},
};

struct vid_std v_stds[] = {
	{"PAL_B", V4L2_STD_PAL_B},
	{"PAL_B1", V4L2_STD_PAL_B1},
	{"PAL_G", V4L2_STD_PAL_G},
	{"PAL_H", V4L2_STD_PAL_H},
	{"PAL_I", V4L2_STD_PAL_I},
	{"PAL_D", V4L2_STD_PAL_D},
	{"PAL_D1", V4L2_STD_PAL_D1},
	{"PAL_K", V4L2_STD_PAL_K},
	{"PAL_M", V4L2_STD_PAL_M},
	{"PAL_N", V4L2_STD_PAL_N},
	{"PAL_Nc", V4L2_STD_PAL_Nc},
	{"PAL_60", V4L2_STD_PAL_60},
	{"NTSC_M", V4L2_STD_NTSC_M},
	{"NTSC_M_JP", V4L2_STD_NTSC_M_JP},
	{"NTSC_443", V4L2_STD_NTSC_443},
	{"NTSC_M_KR", V4L2_STD_NTSC_M_KR},
	{"SECAM_B", V4L2_STD_SECAM_B},
	{"SECAM_D", V4L2_STD_SECAM_D},
	{"SECAM_G", V4L2_STD_SECAM_G},
	{"SECAM_H", V4L2_STD_SECAM_H},
	{"SECAM_K", V4L2_STD_SECAM_K},
	{"SECAM_K1", V4L2_STD_SECAM_K1},
	{"SECAM_L", V4L2_STD_SECAM_L},
	{"SECAM_LC", V4L2_STD_SECAM_LC},
	{"ATSC_8_VSB", V4L2_STD_ATSC_8_VSB},
	{"ATSC_16_VSB", V4L2_STD_ATSC_16_VSB},
	{"NTSC", V4L2_STD_NTSC},
	{"SECAM_DK", V4L2_STD_SECAM_DK},
	{"SECAM", V4L2_STD_SECAM},
	{"PAL_BG", V4L2_STD_PAL_BG},
	{"PAL_DK", V4L2_STD_PAL_DK},
	{"PAL", V4L2_STD_PAL},
	{"B", V4L2_STD_B},
	{"G", V4L2_STD_G},
	{"H", V4L2_STD_H},
	{"L", V4L2_STD_L},
	{"GH", V4L2_STD_GH},
	{"DK", V4L2_STD_DK},
	{"BG", V4L2_STD_BG},
	{"MN", V4L2_STD_MN},
	{"MTS", V4L2_STD_MTS},
	{"525_60", V4L2_STD_525_60},
	{"625_50", V4L2_STD_625_50},
	{"ATSC", V4L2_STD_ATSC},
	{"ALL", V4L2_STD_ALL},
	{"\0", 0},
};

static const char short_options[] =
	"D:d:f:n:a:b:e:g:hi:j:k:l:m:o:p:q:r:t:uv:w:s:z:x:y:";

static const struct option long_options[] = {
	{"Device", required_argument, NULL, 'D'},
	{"drop_frames", required_argument, NULL, 'd'},
	{"save_frames", required_argument, NULL, 's'},
	{"format", required_argument, NULL, 'f'},
	{"name", required_argument, NULL, 'n'},
	{"sd_cam", required_argument, NULL, 'a'},
	{"sd_ipu", required_argument, NULL, 'b'},
	{"gain_grey", required_argument, NULL, 'e'},
	{"gain_red", required_argument, NULL, 'g'},
	{"gain_yellow", required_argument, NULL, 'i'},
	{"gain_cyan", required_argument, NULL, 'j'},
	{"gain_blue", required_argument, NULL, 'k'},
	{"crop_left", required_argument, NULL, 'l'},
	{"crop_top", required_argument, NULL, 'm'},
	{"no_subdev", no_argument, NULL, 'u'},
	{"crop_x", required_argument, NULL, 'o'},
	{"crop_y", required_argument, NULL, 'p'},
	{"binning", required_argument, NULL, 'q'},
	{"flip_hori", required_argument, NULL, 'z'},
	{"flip_vert", required_argument, NULL, 'x'},
	{"reg_file", required_argument, NULL, 'r'},
	{"timeout", required_argument, NULL, 't'},
	{"res_x", required_argument, NULL, 'v'},
	{"res_y", required_argument, NULL, 'w'},
	{"std", required_argument, NULL, 'y'},
	{"help", no_argument, NULL, 'h'},
	{0, 0, 0, 0}
};

/* camera sys path */
static const char *cam_paths[] = {
	"mt9v032",
	"mt9p031",
	NULL
};

static int sri_str_to_fmt(char *fmt_name, struct pm_format *fmt)
{
	uint32_t lv;

	CHECK_VAL(!(fmt_name && fmt), goto done;);
	lv = 0;

	while ('\0' != (pm_formats[lv].name[0])) {
		if (!strcmp(fmt_name, pm_formats[lv].name)) {
			memcpy(fmt, &pm_formats[lv], sizeof(*fmt));
			return lv;
		}

		lv++;
	}

	fprintf(stderr, "%s is not a valid format\n", fmt_name);
 done:
	return -EINVAL;
}

static int sri_str_to_std(char *std_name, struct vid_std *std)
{
	uint32_t lv;

	CHECK_VAL(!(std_name && std), goto done;);

	lv = 0;

	while ('\0' != (v_stds[lv].name[0])) {
		if (!strcmp(std_name, v_stds[lv].name)) {
			memcpy(std->name, v_stds[lv].name, sizeof(*std));
			std->code = v_stds[lv].code;
			return lv;
		}

		lv++;
	}

	fprintf(stderr, "%s is not a valid standard\n", std_name);
 done:
	return -EINVAL;
}

static void sri_print_help(struct sri_args *s, FILE * fp, int argc, char **argv)
{
	uint32_t lv;
	CHECK_VAL(!(s && fp && argv), return;);

	fprintf(fp,
		"Usage: %s [options]\n\n"
		"Options:\n"
		"\n"
		"\t -D [STR]                   video device name [%s]\n"
		"\t -sd_cam [STR]              subdevice cam (make sure this dev is set correctly)[%s]\n"
		"\t -sd_ipu [STR]              subdevice ipu (make sure this dev is set correctly)[%s]\n"
		"\t -no_subdev                 don't use subdevices\n"
		"\n"
		"\t -f [STR]                   format [%s]\n"
		"\n"
		"\t -std                       video standard [%s]\n"
		"\n"
		"\t -r [STR]                   register-file path\n"
		"\n"
		"\t -t [INT]                   application deadline in\n"
		"\t                            seconds [0..255] [%" PRIu8 "]\n"
		"\n"
#ifdef GAIN_BINN_SUPPORT
		"\t -grey_gain [INT]           exclusive for Y8, Y10, Y12:\n"
		"\t                            global gain  [8..1024] [%i]\n"
		"\t -gain_red [INT]            red gain [8..1024] [%i]\n"
		"\t -gain_yellow [INT]         yellow gain [8..1024] [%i]\n"
		"\t -gain_cyan [INT]           cyan gain [8..1024] [%i]\n"
		"\t -gain_blue [INT]           blue gain [8..1024] [%i]\n"
		"\n"
		"\t -binning [INT]             binning [0..1] [%i]\n"
		"\t\n"
		"\t -flip_hori [INT]           flip horizontal [0..1] [%i]\n"
		"\t -flip_vert [INT]           flip vertical [0..1] [%i]\n"
		"\n"
#endif
		"\t -d [INT]                   drop first frames [0..1000][%i]\n"
		"\t -s [INT]                   save n frames [1..100][%i]\n"
		"\n"
		"\t -res_x [INT]               x resolution [48..2592][%i]\n"
		"\t -res_y [INT]               y resolution [48..1944][%i]\n"
		"\n"
#ifdef CROPPING_SUPPORT
		"\t -crop_left [INT]           crop width [0..1000][%i]\n"
		"\t -crop_top [INT]            crop top [0..1000][%i]\n"
		"\t -crop_x [INT]              crop x-res [48..2592][%i]\n"
		"\t -crop_y [INT]              crop y-res [48..1944][%i]\n"
		"\n"
#endif
		"\t -h                         Print this message\n"
		"\n",
		argv[0], s->vdev, s->sdevs.sdev_cam, s->sdevs.sdev_ipu,
		s->format.name, s->v_std.name,
		s->timeout_sec,
#ifdef GAIN_BINN_SUPPORT
		s->sdevs.gain_global, s->sdevs.gain_red,
		s->sdevs.gain_yellow, s->sdevs.gain_cyan,
		s->sdevs.gain_blue, s->sdevs.binning,
		s->sdevs.flip_h, s->sdevs.flip_v,
#endif
		s->drop_frames, s->save_frames, s->sdevs.res_x, s->sdevs.res_y
#ifdef CROPPING_SUPPORT
		,
		s->sdevs.crops[LEFT],
		s->sdevs.crops[TOP],
		s->sdevs.crops[WIDTH], s->sdevs.crops[HEIGHT]
#endif
		);

	fprintf(fp, "\nvalid formats :\n");
	lv = 0;
	while ('\0' != pm_formats[lv].name[0]) {
		fprintf(fp, "[%s] ", pm_formats[lv].name);
		lv++;
	}

	fprintf(fp, "\nvalid video standarts:\n");
	lv = 0;
	while ('\0' != v_stds[lv].name[0]) {
		fprintf(fp, "[%s] ", v_stds[lv].name);
		lv++;
	}

	fprintf(fp, "\n");
}

static int sri_cam_reg_write(uint32_t address, uint64_t reg,
				 uint64_t val, char mode, uint32_t use_subdev)
{
	int ret;
	uint64_t readed = 0;

	if (use_subdev)
		ret = subdev_cam_reg_get(address, reg, &readed);
	else
		ret = videodev_reg_get(address, reg, &readed);

	CHECK_VAL(ret != 0, goto err;);

	switch (mode) {
	case 'a':
		val &= readed;
		break;
	case 'o':
		val |= readed;
		break;
	case 'x':
		val ^= readed;
		break;
	default:
		break;
	}

	if (use_subdev)
		ret = subdev_cam_reg_set(address, reg, val);
	else
		ret = videodev_reg_set(address, reg, val);

	CHECK_VAL(0 != ret, goto err;);

	return 0;
 err:
	return ret;
}

static inline uint32_t sri_int_leng(uint32_t value)
{
	uint32_t count = 0;

	if (!value)
		return 1;

	while (value) {
		value /= 10;
		count++;
	}

	return count;
}

static char *sri_next_filestr(char *save_name, uint32_t numb)
{
	int ret;
	char *resstr;
	uint32_t strl;
	uint32_t sl0;
	uint32_t sl1;
	uint32_t sl2;

	CHECK_VAL(!save_name, ret = EINVAL; goto err;);

	sl0 = strlen(save_name);
	sl1 = sri_int_leng(numb);
	sl2 = strlen(NAME_END);

	strl = sl0 + sl1 + sl2 + 1;

	resstr = malloc(strl * sizeof(*resstr));
	CHECK_VAL(!resstr, ret = ENOMEM; goto err;);

	memcpy(resstr, save_name, sl0);
	sprintf(resstr + sl0, "%u", numb);
	memcpy(resstr + sl0 + sl1, NAME_END, sl2);

	resstr[strl - 1] = '\0';

	return resstr;

 err:
	return ERR_PTR(ret);
}

static int sri_write_to_file(char *save_name, struct image *img,
				 uint32_t img_count)
{
	int ret;
	FILE *fp;
	uint32_t lv;
	char *tmpfname;
	size_t wcount;

	CHECK_VAL(!(save_name && img), ret = EINVAL; goto err;);

	for (lv = 0; lv < img_count; lv++) {
		tmpfname = sri_next_filestr(save_name, lv);
		if (IS_ERR(tmpfname)) {
			ret = PTR_ERR(tmpfname);
			goto err;
		}

		printf("%s\n", tmpfname);

		fp = fopen(tmpfname, "w+");
		CHECK_VAL(!fp, ret = errno; goto err_free;);

		wcount = fwrite(img[lv].data, 1, img[lv].size, fp);
		if (wcount != img[lv].size) {
			ret = errno;
			perror("fwrite");
			goto err_close;
		}

		free(tmpfname);
		fclose(fp);
	}

	return 0;

 err_close:
	fclose(fp);
 err_free:
	free(tmpfname);
 err:
	return ret;
}

static int sri_range_check(struct sri_args *sargs)
{
	int ret;
	bool err;
	struct sub_devices *sds;

	CHECK_VAL(!sargs, ret = EINVAL; goto done;);

	ret = 0;
	err = false;
	sds = &sargs->sdevs;

	//sub device ranges
	if (sds->is_grey) {
		if (!IN_RANGE(sds->gain_global, GAIN_MIN, GAIN_MAX)) {
			fprintf(stderr, "global gain is out of range"
				" [%d..%d]\n", GAIN_MIN, GAIN_MAX);
			err = true;
		}
	} else {
		if (!IN_RANGE(sds->gain_red, GAIN_MIN, GAIN_MAX)) {
			fprintf(stderr, "red gain is out of range"
				" [%d..%d]\n", GAIN_MIN, GAIN_MAX);
			err = true;
		}

		if (!IN_RANGE(sds->gain_yellow, GAIN_MIN, GAIN_MAX)) {
			fprintf(stderr, "yellow gain is out of range"
				" [%d..%d]\n", GAIN_MIN, GAIN_MAX);
			err = true;
		}

		if (!IN_RANGE(sds->gain_cyan, GAIN_MIN, GAIN_MAX)) {
			fprintf(stderr, "cyan gain is out of range"
				" [%d..%d]\n", GAIN_MIN, GAIN_MAX);
			err = true;
		}

		if (!IN_RANGE(sds->gain_blue, GAIN_MIN, GAIN_MAX)) {
			fprintf(stderr, "blue gain is out of range"
				" [%d..%d]\n", GAIN_MIN, GAIN_MAX);
			err = true;
		}
	}

	if (!IS_BOOL(sds->binning)) {
		fprintf(stderr, "binning is not a boolean" " value [0 or 1]\n");
		err = true;
	}

	if (!(IS_BOOL(sds->flip_h) && IS_BOOL(sds->flip_v))) {
		fprintf(stderr, "flips are no boolean value [0 or 1]\n");
		err = true;
	}

	if (!IN_RANGE(sds->res_x, RES_MIN, RES_X_MAX)) {
		fprintf(stderr, "res_x is out of range"
			" [%d..%d]\n", RES_MIN, RES_X_MAX);
		err = true;
	}

	if (!IN_RANGE(sds->res_y, RES_MIN, RES_Y_MAX)) {
		fprintf(stderr, "res_y is out of range"
			" [%d..%d]\n", RES_MIN, RES_Y_MAX);
		err = true;
	}

	if (!IN_RANGE(sds->crops[LEFT], CROP_LT_MIN, CROP_LT_MAX)) {
		fprintf(stderr, "crop_left is out of range"
			" [%d..%d]\n", CROP_LT_MIN, CROP_LT_MAX);
		err = true;
	}

	if (!IN_RANGE(sds->crops[TOP], CROP_LT_MIN, CROP_LT_MAX)) {
		fprintf(stderr, "crop_top is out of range"
			" [%d..%d]\n", CROP_LT_MIN, CROP_LT_MAX);
		err = true;
	}

	if (!IN_RANGE(sds->crops[WIDTH], CROP_XY_MIN, CROP_X_MAX)) {
		fprintf(stderr, "crop_x is out of range"
			" [%d..%d]\n", CROP_XY_MIN, CROP_X_MAX);
		err = true;
	}

	if (!IN_RANGE(sds->crops[HEIGHT], CROP_XY_MIN, CROP_Y_MAX)) {
		fprintf(stderr, "crop_y is out of range"
			" [%d..%d]\n", CROP_XY_MIN, CROP_Y_MAX);
		err = true;
	}

	if (!IN_RANGE(sargs->drop_frames, DROP_FR_MIN, DROP_FR_MAX)) {
		fprintf(stderr, "drop_frames is out of range"
			" [%d..%d]\n", DROP_FR_MIN, DROP_FR_MAX);
		err = true;
	}

	if (!IN_RANGE(sargs->save_frames, SAVE_FR_MIN, SAVE_FR_MAX)) {
		fprintf(stderr, "setted save_frames is out of range"
			" [%d..%d]\n", SAVE_FR_MIN, SAVE_FR_MAX);
		err = true;
	}

	if (err)
		ret = ERANGE;
 done:
	return ret;
}

static struct sri_args *sri_args_get(int argc, char **argv)
{
	int ret;
	int optindex;
	int opt;

	struct sri_args *sargs;

	CHECK_VAL((!argv), ret = EINVAL; goto err0;);

	sargs = calloc(1, sizeof(*sargs));
	CHECK_VAL(!sargs, ret = ENOMEM; goto err0;);

	/* set defaults */
	strcpy(sargs->vdev, DEV_VID_DEF);	//video_device
	strcpy(sargs->sdevs.sdev_cam, SDEV_CAM_DEF);	//subdevice cam
	strcpy(sargs->sdevs.sdev_ipu, SDEV_IPU_DEF);	//subdevice ipu

	ret = sri_str_to_fmt(FMT_DEF, &sargs->format);
	CHECK_VAL(ret < 0, ret -= ret; goto err0;);

	ret = sri_str_to_std(STD_DEF, &sargs->v_std);
	CHECK_VAL(ret < 0, ret -= ret; goto err0;);

	sargs->sdevs.is_grey = true;

	sargs->sdevs.gain_global = GAIN_DEF;
	sargs->sdevs.gain_red = GAIN_DEF;
	sargs->sdevs.gain_yellow = GAIN_DEF;
	sargs->sdevs.gain_cyan = GAIN_DEF;
	sargs->sdevs.gain_blue = GAIN_DEF;

	sargs->timeout_sec = 0;

	sargs->sdevs.binning = false;
	sargs->sdevs.flip_h = 0;
	sargs->sdevs.flip_v = 0;
	sargs->sdevs.res_x = 2592;
	sargs->sdevs.res_y = 1944;
	sargs->sdevs.crops[LEFT] = 16;
	sargs->sdevs.crops[TOP] = 54;
	sargs->sdevs.crops[WIDTH] = sargs->sdevs.res_x;
	sargs->sdevs.crops[HEIGHT] = sargs->sdevs.res_y;
	sargs->sdevs.use_subdev = true;

	sargs->drop_frames = 2;
	sargs->save_frames = 1;
	strcpy(sargs->name_fmt, NAME_FMT_DEF);

	sargs->buffs = NULL;
	sargs->set_reg = false;
	sargs->set_std = false;

	while (1) {
		optindex = -1;
		opt = -1;

		opt = getopt_long_only(argc, argv, short_options, long_options,
					   &optindex);

		if (-1 == opt)
			break;

		switch (opt) {
		case 0:
			break;
		case 'D':
			strcpy(sargs->vdev, optarg);
			break;
		case 'f':
			ret = sri_str_to_fmt(optarg, &sargs->format);
			CHECK_VAL(ret < 0, ret = -ret; goto err1;);

			if (ret > 2)
				sargs->sdevs.is_grey = false;
			break;
		case 'y':
			ret = sri_str_to_std(optarg, &sargs->v_std);
			CHECK_VAL(ret < 0, ret = -ret; goto err1;);
			sargs->set_std = true;
			break;
		case 'd':
			sargs->drop_frames =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 's':
			sargs->save_frames =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 't':
			sargs->timeout_sec =
				(uint8_t) strtoul(optarg, NULL, 10);
			break;
		case 'n':
			strcpy(sargs->name_fmt, optarg);
			break;
		case 'a':
			strcpy(sargs->sdevs.sdev_cam, optarg);
			break;
		case 'b':
			strcpy(sargs->sdevs.sdev_ipu, optarg);
			break;
		case 'e':
			sargs->sdevs.gain_global =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 'g':
			sargs->sdevs.gain_red =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 'i':
			sargs->sdevs.gain_yellow =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 'j':
			sargs->sdevs.gain_cyan =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 'k':
			sargs->sdevs.gain_blue =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 'l':
			sargs->sdevs.apply_crop = true;
			sargs->sdevs.crops[LEFT] =
				(uint32_t) strtoul(optarg, NULL, 10);
			break;
		case 'm':
			sargs->sdevs.apply_crop = true;
			sargs->sdevs.crops[TOP] =
				(int32_t) strtoul(optarg, NULL, 10);
			break;
		case 'o':
			sargs->sdevs.apply_crop = true;
			sargs->sdevs.crops[WIDTH] =
				(int32_t) strtol(optarg, NULL, 10);
			break;
		case 'p':
			sargs->sdevs.apply_crop = true;
			sargs->sdevs.crops[HEIGHT] =
				(int32_t) strtol(optarg, NULL, 10);
			break;
		case 'q':
			sargs->sdevs.binning =
				(int32_t) strtol(optarg, NULL, 10);
			break;
		case 'z':
			sargs->sdevs.flip_h =
				(int32_t) strtol(optarg, NULL, 10);
			break;
		case 'x':
			sargs->sdevs.flip_v =
				(int32_t) strtol(optarg, NULL, 10);
			break;
		case 'v':
			sargs->sdevs.res_x = (int32_t) strtol(optarg, NULL, 10);
			break;
		case 'w':
			sargs->sdevs.res_y = (int32_t) strtol(optarg, NULL, 10);
			break;
		case 'r':
			strcpy(sargs->reg_file, optarg);
			sargs->set_reg = true;
			break;
		case 'u':
			sargs->sdevs.use_subdev = false;
			break;
		case 'h':
		default:
			sri_print_help(sargs, stdout, argc, argv);
			ret = 0;
			goto err0;
		}
	}

	ret = sri_range_check(sargs);
	if (ret != 0)
		goto err1;

	return sargs;
 err1:
	free(sargs);
 err0:
	return ERR_PTR(ret);
}

static void sri_print_args(struct sri_args *sargs)
{
	CHECK_VAL(!sargs, return;);

	printf("video device =		%s\n", sargs->vdev);
	printf("cam subdev =		%s\n", sargs->sdevs.sdev_cam);
	printf("ipu subdev =		%s\n", sargs->sdevs.sdev_ipu);
	printf("use subdevs = 		%s\n\n", sargs->sdevs.use_subdev ? "true" : "false");
	printf("timeout in secs =	%" PRIu8 "\n\n", sargs->timeout_sec);

#ifdef GAIN_BINN_SUPPORT
	if (sargs->sdevs.is_grey)
		printf("global gain =		" PRIi32 "\n",
			   sargs->sdevs.gain_global);
	else {
		printf("gain red =		%" PRIu32 "\n",
			   sargs->sdevs.gain_red);
		printf("gain yellow =		%" PRIu32 "\n",
			   sargs->sdevs.gain_yellow);
		printf("gain cyan =		%" PRIu32 "\n",
			   sargs->sdevs.gain_cyan);
		printf("gain blue =		%" PRIu32 "\n",
			   sargs->sdevs.gain_blue);
	}

	printf("binning =		%d\n\n", sargs->sdevs.binning);
#endif

	if (INT32_MAX == sargs->sdevs.res_x)
		printf("res size x =	 MAX\n");
	if (INT32_MAX == sargs->sdevs.res_y)
		printf("res size x =	 MAX\n");

	printf("res size x =		%" PRIi32 "\n", sargs->sdevs.res_x);
	printf("res size y =		%" PRIi32 "\n\n", sargs->sdevs.res_y);
#ifdef CROPPING_SUPPORT
	printf("apply crop =		%s\n", sargs->sdevs.apply_crop ? "true" : "false");
	printf("crop left =		%" PRIi32 "\n", sargs->sdevs.crops[LEFT]);
	printf("crop top =		%" PRIi32 "\n", sargs->sdevs.crops[TOP]);
	printf("crop width (x) =	%" PRIi32 "\n", sargs->sdevs.crops[WIDTH]);
	printf("crop height (y) =	%" PRIi32 "\n", sargs->sdevs.crops[HEIGHT]);

	printf("flip horiz =		%" PRIi32 "\n", sargs->sdevs.flip_h);
	printf("glip vert  =		%" PRIi32 "\n\n", sargs->sdevs.flip_v);
#endif
	/* setting standards not supported */
	printf("format =		%s (pixf = 0x%x mbusf = 0x%x)\n",
		   sargs->format.name, sargs->format.pfmt_code,
		   sargs->format.mfmt_code);
	if (sargs->set_reg)
		printf("register path =	%s\n", sargs->reg_file);

	printf("save name fmt =		%sX%s\n", sargs->name_fmt, NAME_END);
	printf("save frames =		%" PRIu32 "\n", sargs->save_frames);
	printf("drop frames =		%" PRIu32 "\n", sargs->drop_frames);
	printf("\n");
}

static int sri_subdevs_init(struct sri_args *sargs)
{
	int ret;
	struct sub_devices *sdevs;

	CHECK_VAL(!sargs, ret = EINVAL; goto err0;);

	if (!sargs->sdevs.use_subdev)
		return 0;

	ret = 0;
	sdevs = &sargs->sdevs;

	ret = subdev_cam_init(sdevs->sdev_cam);
	CHECK_VAL(ret != 0, goto err0;);

	ret = subdev_ipu_init(sdevs->sdev_ipu);
	CHECK_VAL(ret != 0, goto err1;);

	ret = subdev_cam_format_set(0, &sdevs->res_x,
					&sdevs->res_y, &sargs->format.mfmt_code);
	CHECK_VAL(ret != 0, goto err2;);

	if (sdevs->apply_crop) {
		ret = subdev_cam_format_set_crop(0,
						&sdevs->crops[LEFT],
						&sdevs->crops[TOP],
						&sdevs->crops[WIDTH],
						&sdevs->crops[HEIGHT]);
		CHECK_VAL(ret != 0, goto err2;);
	}

	ret = subdev_ipu_format_set(0, &sdevs->res_x,
					&sdevs->res_y, &sargs->format.mfmt_code);
	CHECK_VAL(ret != 0, goto err2;);

	ret = subdev_ipu_format_set(1, &sdevs->res_x,
					&sdevs->res_y, &sargs->format.mfmt_code);
	CHECK_VAL(ret != 0, goto err2;);

	return 0;
 err2:
	subdev_ipu_delete();
 err1:
	subdev_cam_delete();
 err0:
	return ret;
}

static int sri_camera_check()
{
	int ret;
	struct stat st;
	char *tmpp;
	const char **tmppaths;
	size_t sleng;
	size_t pathleng;
	bool camfound;

	CLEAR(st);
	pathleng = strlen(I2C_PATH);
	camfound = false;
	tmppaths = cam_paths;

	while (*tmppaths && !camfound) {
		sleng = strlen(*tmppaths);

		tmpp = malloc(sleng + pathleng + 1);
		if (!tmpp) {
			ret = ENOMEM;
			goto done;
		}

		memcpy(tmpp, I2C_PATH, pathleng);
		memcpy(tmpp + pathleng, *tmppaths, sleng);
		tmpp[pathleng + sleng] = '\0';

		ret = stat(tmpp, &st);
		free(tmpp);
		if (ret != 0) {
			if (errno != ENOENT) {
				ret = errno;
				perror("stat");
				goto done;
			}
			tmppaths++;
		} else {
			ret = 0;
			printf("found camera: %s\n", *tmppaths);
			camfound = true;
		}

	}

	ret = 0;
 done:
	return ret;
}

int main(int argc, char **argv)
{

	int ret;
	uint32_t img_size;
	uint32_t lv;
	struct sri_args *sargs;
	struct image *images;

	sargs = sri_args_get(argc, argv);
	if (IS_ERR(sargs)) {
		ret = PTR_ERR(sargs);
		goto done;
	}

	ret = sri_camera_check();
	if (ret != 0) {
		fprintf(stderr,
			"Camera not found please check the hardware and kernel log\n");
		goto fargs;
	}

	sri_print_args(sargs);

	ret = videodev_init(sargs->vdev);
	CHECK_VAL(ret != 0, goto fargs;);

	ret = sri_subdevs_init(sargs);
	CHECK_VAL(ret != 0, goto fargs;);

	ret = videodev_pfmt_set(sargs->format.pfmt_code,
				&sargs->sdevs.res_x, &sargs->sdevs.res_y);
	CHECK_VAL(ret != 0, goto del_sdevs;);

	if (sargs->set_std) {
		ret = videodev_std_set(sargs->v_std.code);
		if (0 != ret) {
			fprintf(stderr, "unable to change video device\n");
			videodev_stds_print();	//ret not necessary (already error)
			goto del_vdev;
		}
	}

	if (sargs->set_reg) {
		ret = reg_file_parse(sargs->reg_file,
					 sri_cam_reg_write,sargs->sdevs.use_subdev);
		CHECK_VAL(ret != 0, goto del_vdev;);
	}

	ret = videodev_capture(sargs->save_frames,
				   sargs->drop_frames, sargs->timeout_sec, &images,
				   &img_size);
	CHECK_VAL(ret != 0, goto del_vdev;);

	ret = sri_write_to_file(sargs->name_fmt, images, img_size);

	for (lv = 0; lv < img_size; lv++)
		free(images[lv].data);
	free(images);

	ret = 0;
 del_vdev:
	videodev_delete();
 del_sdevs:
	subdev_cam_delete();
	subdev_ipu_delete();
 fargs:
	free(sargs);
 done:
	exit(ret);
}
