/*
 * Copyright (C) 2015 PHYTEC Messtechnik GmbH,
 * Author:   Alexander Fratzer <a.fratzer@phytec.de>
 *           Alexander Aring   <a.aring@phytec.de>
 * Released under the MIT license
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include <inttypes.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>

#include <linux/videodev2.h>
#include <linux/v4l2-subdev.h>

#include "device_utils.h"
#include "videodev.h"
#include "errhandl.h"

enum sub_device { CAM, IPU };

static int video_fd = -1;

static int subdev_fds[] = { -1, -1 };

/**
 * @brief Setup the videodevice for capturing. Checks availabilty
 * of streaming and capturing.
 *
 * @param video_fd - filedescriptor of (open) video-device
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int videodev_setup_capture();

/**
 * @brief Maps the kernelspace into the buffers (see spec) 
 *
 * @param video_fd - filedescriptor of (open) video-device
 * @param bufcount - resulting number of buffers on success
 *
 * @return - success: pointer to the buffer array.
 *		   fail: ERRORCODE (cast to pointer)
 */
static struct buffer *videodev_mmap_init(uint32_t * bufcount);

/**
 * @brief Unmaps the mapped buffers
 *
 * @param buffs - mapped buffers (in array)
 * @param buf_count - number if elements in buffs
 */
static void videodev_unmap(struct buffer *buffs, uint32_t buf_count);

/**
 * @brief Start capturing
 *
 * @param video_fd - filedescriptior of (open) video-device
 * @param buf_count - number of buffers
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int videodev_capture_start(uint32_t buf_count);

/**
 * @brief Stop the capturing process.
 *
 * @param video_fd - filedescriptor of (open) video-device
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int videodev_capture_stop();

/**
 * @brief Get data from video-driver and copies it to image.
 *
 * This function allocates memory (image - data). It is in 
 * the responsibilty of the user to free these memory. The
 * image structe has to be pre-allocated.
 *
 * A call of this only works if the process of capturing
 * is enabled, the buffers are mapped and the calling 
 * process got response (via select).
 * 
 * If NULL == buf_count or image, the function drops
 * the image
 *
 * @param video_fd - filedescriptor of (open) video-device
 * @param mapped_bufs - mapped buffer-array
 * @param buf_count - number of buffers
 * @param img - resulting image
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int videodev_frame_op(struct buffer *mapped_bufs,
				 uint32_t buf_count, struct image *img);

/**
 * @brief The Process of streaming. This function gets data from
 * the driver (using the mapped buffer) and copies these data
 * from to image.
 *
 * @param video_fd - filedescriptor of (open) video-device
 * @param mapped_bufs - mapped buffer-array
 * @param buf_count - number of elements in mapped_bufs
 * @param saves - number of images for saving
 * @param drops - number if images drops before saving
 * @param timeout_sec - max time (seconds) waiting for driver
 *					  driver response
 * @param res_img - resulting array of images
 * @param ri_size - resulting size of res_img 
 *
 * @return - sucess: 0
 *		   fail: ERRORCODE
 */
static int videodev_stream(struct buffer *mapped_bufs,
			   uint32_t buf_count, uint32_t saves, uint32_t drops,
			   uint8_t timeout_sec, struct image **res_img,
			   uint32_t * ri_size);

/**
 * @brief Wrapper subdevice init
 *
 * @param devname - path to device name
 * @param type - subdevice type
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int subdev_init(char *devname, enum sub_device type);

/**
 * @brief Wrapper subdevice uninit
 *
 * @param type - subdevice type
 */
static void subdev_delete(enum sub_device type);

/**
 * @brief Wrapper format setter
 *
 * @param pad - pad
 * @param width - resulting x
 * @param height - resulting y
 * @param fmt_code - pixel format
 * @param type - subdeivce type
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int subdev_format_set(int32_t pad, int32_t * width, int32_t * height,
				 enum v4l2_mbus_pixelcode *fmt_code,
				 enum sub_device type);

/**
 * @brief - Wrapper extctrl set
 *
 * @param id - subdev id
 * @param value - value for driver
 * @param type - subdevice type
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int subdev_extctrl_set(uint32_t id, int32_t value, enum sub_device type);

/**
 * @brief - Wrapper ctrl set
 *
 * @param id - subdev id
 * @param value - value for driver
 * @param type - subdevice type
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int subdev_ctrl_set(uint32_t id, int32_t value, enum sub_device type);

/**
 * @brief Wrapper Hardset register function 
 *
 * @param address - i2c address
 * @param reg - register address
 * @param data - data
 * @param type - subdevice type
 *
 * @return - success: 0
 *		   fail: ERRORCODE
 */
static int subdev_reg_set(uint32_t address, uint64_t reg,
			  uint64_t data, enum sub_device type);

/**
 * @brief - Wrapper Hardget register function
 *
 * @param address - i2c address
 * @param reg - register address
 * @param data - resulting data
 * @param type - subdevice type
 *
 * @return - success: 0
 */
static int subdev_reg_get(uint32_t address, uint64_t reg,
			  uint64_t * data, enum sub_device type);

static int subdev_init(char *devname, enum sub_device type)
{
	int ret;

	CHECK_VAL(!devname, ret = EINVAL; goto done;);

	if (subdev_fds[type] != -1) {
		ret = EBUSY;
		fprintf(stderr, "sub_device %s already in use!\n", devname);
		goto done;
	}

	subdev_fds[type] = open(devname, O_RDWR | O_NONBLOCK);
	if (-1 == subdev_fds[type]) {
		ret = errno;
		perror("subdev_cam_init: open");
		goto done;
	}

	ret = 0;
 done:
	return ret;
}

int subdev_cam_init(char *devname)
{
	CHECK_VAL(!devname, return EINVAL;);

	return subdev_init(devname, CAM);
}

int subdev_ipu_init(char *devname)
{
	CHECK_VAL(!devname, return EINVAL;);

	return subdev_init(devname, IPU);
}

void subdev_cam_delete()
{
	subdev_delete(CAM);
}

void subdev_ipu_delete()
{
	subdev_delete(IPU);
}

static void subdev_delete(enum sub_device type)
{
	if (-1 == subdev_fds[type]) {
		fprintf(stderr, "sub_device is not in use\n");
		goto done;
	}

	if (-1 == close(subdev_fds[type])) {
		perror("subdev close");
		goto done;
	}

	subdev_fds[type] = -1;

 done:
	return;
}

int subdev_cam_reg_set(uint32_t address, uint64_t reg, uint64_t data)
{
	return subdev_reg_set(address, reg, data, CAM);
}

int subdev_ipu_reg_set(uint32_t address, uint64_t reg, uint64_t data)
{
	return subdev_reg_set(address, reg, data, IPU);
}

int subdev_cam_reg_get(uint32_t address, uint64_t reg, uint64_t * data)
{
	CHECK_VAL(!data, return EINVAL;);

	return subdev_reg_get(address, reg, data, CAM);
}

int subdev_ipu_reg_get(uint32_t address, uint64_t reg, uint64_t * data)
{
	CHECK_VAL(!data, return EINVAL;);

	return subdev_reg_get(address, reg, data, IPU);
}

int subdev_cam_format_set(int32_t pad, int32_t * resx, int32_t * resy,
			  enum v4l2_mbus_pixelcode *fmt_code)
{
	CHECK_VAL(!(resx && resy && fmt_code), return EINVAL;);

	return subdev_format_set(pad, resx, resy, fmt_code, CAM);
}

int subdev_cam_format_set_crop(int32_t pad,
			  int32_t *crop_left, int32_t *crop_top,
			  int32_t *crop_width, int32_t *crop_height)
{
	int ret;
	struct v4l2_subdev_crop crop;

	CLEAR(crop);
	crop.pad = pad;
	crop.which = V4L2_SUBDEV_FORMAT_ACTIVE;
	crop.rect.left = *crop_left; /* __s32 */
	crop.rect.top = *crop_top;
	crop.rect.width = *crop_width; /* __u32 */
	crop.rect.height = *crop_height;

	/*
	 * TODO
	 *
	 * This is an Obsolete API Elements interface and may be removed in the
	 * future. It is superseded by the selection API.
	 *
	 *     http://static.lwn.net/kerneldoc/media/uapi/v4l/vidioc-subdev-g-crop.html#vidioc-subdev-g-crop
	 *
	 * Use VIDIOC_SUBDEV_s_SELECTION instead.
	 */
	ret = xioctl(subdev_fds[CAM], VIDIOC_SUBDEV_S_CROP, &crop);
	if (ret < 0) {
		fprintf(stderr, "Error: Cannot set crop for dev: %s\n",
			strerror(errno));
		return -1;
	}

	if (*crop_left != crop.rect.left)
		fprintf(stderr, "Warning: Adjust 'crop left' from %d to %d!\n",
			*crop_left, crop.rect.left);
	if (*crop_top != crop.rect.top)
		fprintf(stderr, "Warning: Adjust 'crop top' from %d to %d!\n",
			*crop_top, crop.rect.top);
	if (*crop_width != crop.rect.width)
		fprintf(stderr, "Warning: Adjust 'crop width' from %u to %u!\n",
			*crop_width, crop.rect.width);
	if (*crop_height != crop.rect.height)
		fprintf(stderr, "Warning: Adjust 'crop height' from %u to %u!\n",
			*crop_height, crop.rect.height);

	*crop_left = crop.rect.left;
	*crop_top = crop.rect.top;
	*crop_width = crop.rect.width;
	*crop_height = crop.rect.height;

	return 0;
}

int subdev_ipu_format_set(int32_t pad, int32_t * resx, int32_t * resy,
			  enum v4l2_mbus_pixelcode *fmt_code)
{
	CHECK_VAL(!(resx && resy && fmt_code), return EINVAL;);

	return subdev_format_set(pad, resx, resy, fmt_code, IPU);
}

int subdev_cam_extctrl_set(uint32_t id, int32_t value)
{
	return subdev_extctrl_set(id, value, CAM);
}

int subdev_ipu_extctrl_set(uint32_t id, int32_t value)
{

	return subdev_extctrl_set(id, value, IPU);
}

int subdev_cam_ctrl_set(uint32_t id, int32_t value)
{
	return subdev_ctrl_set(id, value, CAM);
}

int subdev_ipu_ctrl_set(uint32_t id, int32_t value)
{
	return subdev_ctrl_set(id, value, IPU);
}

int subdev_format_set(int32_t pad, int32_t * width, int32_t * height,
			  enum v4l2_mbus_pixelcode *fmt_code, enum sub_device type)
{
	int ret;
	int sfd;
	struct v4l2_subdev_format sdfmt;

	CHECK_VAL(!(width && height && fmt_code), ret = EINVAL; goto done;);

	CLEAR(sdfmt);

	sfd = subdev_fds[type];

	sdfmt.pad = pad;
	sdfmt.which = V4L2_SUBDEV_FORMAT_ACTIVE;
	sdfmt.format.width = *width;
	sdfmt.format.height = *height;
	sdfmt.format.code = *fmt_code;

	ret = xioctl(sfd, VIDIOC_SUBDEV_S_FMT, &sdfmt);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_SUBDEV_S_FMT");
		goto done;
	}

	if (*width != sdfmt.format.width || *height != sdfmt.format.height) {
		fprintf(stderr,
			"Warning: pixel format does not support width and "
			"height. Pixel width and hight is changed to:\n"
			"res_x = %d\nres_y = %d\n",
			sdfmt.format.width, sdfmt.format.height);

		*width = sdfmt.format.width;
		*height = sdfmt.format.height;
	}

	ret = 0;
 done:
	return ret;
}

int subdev_extctrl_set(uint32_t id, int32_t value, enum sub_device type)
{
	int ret;
	struct v4l2_ext_control ctrl;
	struct v4l2_ext_controls ctrls;

	CLEAR(ctrl);
	CLEAR(ctrls);

	ctrl.id = id;
	ctrl.size = 0;
	ctrl.value = value;
	ctrl.value64 = value;

	ctrls.ctrl_class = V4L2_CTRL_CLASS_CAMERA;
	ctrls.count = 1;
	ctrls.controls = &ctrl;

	ret = xioctl(subdev_fds[type], VIDIOC_S_EXT_CTRLS, &ctrls);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_S_EXT_CTRLS");
		goto done;
	}

	ret = 0;
 done:
	return ret;
}

int subdev_ctrl_set(uint32_t id, int32_t value, enum sub_device type)
{
	int ret;
	struct v4l2_control ctl;

	CLEAR(ctl);

	ctl.id = id;
	ctl.value = value;

	ret = xioctl(subdev_fds[type], VIDIOC_S_CTRL, &ctl);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_S_CTRL");
		goto done;
	}

	ret = 0;
 done:
	return ret;
}

int subdev_reg_set(uint32_t address, uint64_t reg,
		   uint64_t data, enum sub_device type)
{
	int ret;
	struct v4l2_dbg_register camreg;
	struct v4l2_dbg_match regmatch;

	CLEAR(camreg);
	CLEAR(regmatch);

	regmatch.type = V4L2_CHIP_MATCH_SUBDEV;
	regmatch.addr = address;

	camreg.match = regmatch;
	camreg.size = REGISTER_SIZE;
	camreg.reg = reg;
	camreg.val = data;

	ret = xioctl(subdev_fds[type], VIDIOC_DBG_S_REGISTER, &camreg);
	if (ret < 0) {
		ret = errno;
		perror("VIDIOC_DBG_S_REGISTER");
		goto done;
	}

	ret = 0;
 done:
	return ret;
}

int subdev_reg_get(uint32_t address, uint64_t reg,
		   uint64_t * data, enum sub_device type)
{
	int ret;
	struct v4l2_dbg_register camreg;
	struct v4l2_dbg_match regmatch;

	CHECK_VAL(!data, ret = EINVAL; goto done;);

	CLEAR(camreg);
	CLEAR(regmatch);

	regmatch.type = V4L2_CHIP_MATCH_SUBDEV;
	regmatch.addr = address;

	camreg.match = regmatch;
	camreg.size = REGISTER_SIZE;
	camreg.reg = reg;
	camreg.val = 0;

	ret = xioctl(subdev_fds[type], VIDIOC_DBG_G_REGISTER, &camreg);
	if (ret < 0) {
		ret = errno;
		perror("VIDIOC_DBG_G_REGISTER");
		goto done;
	}

	*data = camreg.val;

	ret = 0;
 done:
	return ret;
}

int videodev_pfmt_set(uint32_t pix_fmt, int32_t * resx, int32_t * resy)
{
	int ret;
	struct v4l2_format fmt;

	CHECK_VAL(!(resx && resy), ret = EINVAL; goto done;);

	CLEAR(fmt);

	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.width = *resx;
	fmt.fmt.pix.height = *resy;
	fmt.fmt.pix.pixelformat = pix_fmt;
	fmt.fmt.pix.field = V4L2_FIELD_ANY;

	ret = xioctl(video_fd, VIDIOC_S_FMT, &fmt);
	if (-1 == ret) {
		if (EINVAL == errno) {
			fprintf(stderr,
				"video does not support choosen pixel format\n");
			ret = EPERM;
		} else {
			ret = errno;
			perror("VIDIOC_S_FMT");
		}
		goto done;
	}

	if (*resx != fmt.fmt.pix.width || *resy != fmt.fmt.pix.height) {
		fprintf(stderr,
			"Warning: pixel format does not support width and "
			"height. Pixel width and height is changed to:\n"
			"res_x = %d\nres_y = %d\n",
			fmt.fmt.pix.width, fmt.fmt.pix.height);

		*resx = fmt.fmt.pix.width;
		*resy = fmt.fmt.pix.height;
	}

	ret = 0;
 done:
	return ret;
}

int videodev_reg_set(uint32_t address, uint64_t reg,
		   uint64_t data)
{
	int ret;
	struct v4l2_dbg_register camreg;
	struct v4l2_dbg_match regmatch;

	CLEAR(camreg);
	CLEAR(regmatch);

	regmatch.type = V4L2_CHIP_MATCH_SUBDEV;
	regmatch.addr = address;

	camreg.match = regmatch;
	camreg.size = REGISTER_SIZE;
	camreg.reg = reg;
	camreg.val = data;

	ret = xioctl(video_fd, VIDIOC_DBG_S_REGISTER, &camreg);
	if (ret < 0) {
		ret = errno;
		perror("VIDIOC_DBG_S_REGISTER");
		goto done;
	}

	ret = 0;
 done:
	return ret;
}

int videodev_reg_get(uint32_t address, uint64_t reg,
		   uint64_t * data)
{
	int ret;
	struct v4l2_dbg_register camreg;
	struct v4l2_dbg_match regmatch;

	CHECK_VAL(!data, ret = EINVAL; goto done;);

	CLEAR(camreg);
	CLEAR(regmatch);

	regmatch.type = V4L2_CHIP_MATCH_SUBDEV;
	regmatch.addr = address;

	camreg.match = regmatch;
	camreg.size = REGISTER_SIZE;
	camreg.reg = reg;
	camreg.val = 0;

	ret = xioctl(video_fd, VIDIOC_DBG_G_REGISTER, &camreg);
	if (ret < 0) {
		ret = errno;
		perror("VIDIOC_DBG_G_REGISTER");
		goto done;
	}

	*data = camreg.val;

	ret = 0;
 done:
	return ret;
}


int videodev_init(char *devname)
{
	int ret;

	CHECK_VAL(!devname, ret = EINVAL; goto done;);

	if (-1 != video_fd) {
		ret = EBUSY;
		fprintf(stderr, "video device is currently in use\n");
		goto done;
	}

	video_fd = open(devname, O_RDWR | O_NONBLOCK);
	if (-1 == video_fd) {
		ret = errno;
		perror("videodev_init: open");
		goto done;
	}

	ret = 0;
 done:
	return ret;
}

void videodev_delete()
{
	if (-1 == video_fd) {
		fprintf(stderr, "videodevice is not in use\n");
		goto done;
	}

	if (-1 == close(video_fd)) {
		perror("videodev_delete: close");
		goto done;
	}

	video_fd = -1;

 done:
	return;
}

static int videodev_setup_capture()
{
	int ret;
	struct v4l2_capability cap;
	struct v4l2_cropcap cropcap;
	struct v4l2_crop crop;

	CLEAR(cap);
	CLEAR(cropcap);
	CLEAR(crop);

	ret = xioctl(video_fd, VIDIOC_QUERYCAP, &cap);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_QUERYCAP");
		goto err;
	}

	if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
		fprintf(stderr, "Videodevice does not support capturing\n"
			"Are you sure correct video device is selected?\n");
		ret = EPERM;
		goto err;
	}

	if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
		fprintf(stderr, "Videodevice does not support streaming\n");
		ret = EPERM;
		goto err;
	}

/*
 * I don't think this code is still needed. Setting the crop factor on the
 * device /dev/videoX always returned "Videodevice does not support croping".
 * Furtheremore the crop factors must be set on the camera subdevice. See
 * function subdev_cam_format_set_crop.
 *
 * TODO remove this code.
 */
#if 0
	cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	ret = xioctl(video_fd, VIDIOC_CROPCAP, &cropcap);	//ignoreable
	if (-1 != ret) {

		crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		crop.c = cropcap.defrect;	//reset

		ret = xioctl(video_fd, VIDIOC_S_CROP, &crop);	//not necessary for functionality
		if (-1 == ret) {
			/* ignoreble not necessary for functionality and atm not supported */
		}
	} else {
		printf("Info: Videodevice does not support croping\n");
	}
#endif

	return 0;
 err:
	return ret;
}

static struct buffer *videodev_mmap_init(uint32_t * bufcount)
{
	int ret;
	int32_t lv;
	struct buffer *res;
	struct v4l2_requestbuffers req;
	struct v4l2_buffer buf;

	CHECK_VAL(!bufcount, ret = EINVAL; goto err;);

	CLEAR(req);

	req.count = 2;		//one displayed + one filled by application
	req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_MMAP;

	ret = xioctl(video_fd, VIDIOC_REQBUFS, &req);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_REQBUFS");
		goto err;
	}

	if (req.count < 2) {
		fprintf(stderr, "Not enough memory for buffers\n");
		ret = ENOMEM;
		goto err;
	}

	res = (struct buffer *)calloc(req.count, sizeof(*res));
	if (!res) {
		ret = ENOMEM;
		goto err;
	}

	for (lv = 0; lv < req.count; lv++) {
		CLEAR(buf);

		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = lv;

		ret = xioctl(video_fd, VIDIOC_QUERYBUF, &buf);
		if (-1 == ret) {
			ret = errno;
			perror("VIDIOC_QUERYBUF");
			goto err_free;
		}

		res[lv].leng = buf.length;
		res[lv].start = mmap(NULL, buf.length, PROT_READ | PROT_WRITE,
					 MAP_SHARED, video_fd, buf.m.offset);
		if (MAP_FAILED == res[lv].start) {
			ret = errno;
			perror("mmap");
			goto err_unmap;
		}
	}

	*bufcount = req.count;

	return res;
 err_unmap:
	while (lv >= 0) {
		munmap(res[lv].start, res[lv].leng);
		lv--;
	}
 err_free:
	free(res);
 err:
	return ERR_PTR(ret);
}

static void videodev_unmap(struct buffer *buffs, uint32_t buf_count)
{
	int32_t lv;

	CHECK_VAL(!buffs, return;);

	for (lv = 0; lv < buf_count; lv++)
		munmap(buffs[lv].start, buffs[lv].leng);

	free(buffs);
}

static int videodev_capture_start(uint32_t buf_count)
{
	int ret;
	uint32_t lv;
	enum v4l2_buf_type bufftype;
	struct v4l2_buffer buf;

	for (lv = 0; lv < buf_count; lv++) {
		CLEAR(buf);
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = lv;

		ret = xioctl(video_fd, VIDIOC_QBUF, &buf);
		if (-1 == ret) {
			ret = errno;
			perror("VIDIOC_QBUF");
			goto err;
		}
	}

	bufftype = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	ret = xioctl(video_fd, VIDIOC_STREAMON, &bufftype);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_STREAMON");
		fprintf(stderr, "Chosen pixel-format is not supported\n");
		goto err;
	}

	return 0;
 err:
	return ret;

}

static int videodev_frame_op(struct buffer *mapped_bufs,
				 uint32_t buf_count, struct image *img)
{
	int ret;
	int tmpret;
	struct v4l2_buffer buf;

	CHECK_VAL(!(buf_count), ret = EINVAL; goto err;);

	CLEAR(buf);
	tmpret = 0;

	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;

	ret = xioctl(video_fd, VIDIOC_DQBUF, &buf);
	if (-1 == ret) {
		ret = errno;
		if (ret != EAGAIN)
			perror("VIDIOC_DQBUF");

		goto err;
	}

	assert(buf.index < buf_count);

	if (mapped_bufs && img) {
		img->data = (uint8_t *) malloc(buf.bytesused);
		printf("Info: %u Bytes\n", buf.bytesused);
		if (img->data) {
			memcpy(img->data, mapped_bufs[buf.index].start,
				   buf.bytesused);
			img->size = buf.bytesused;
			tmpret = 0;
		} else {
			tmpret = ENOMEM;
		}
	}

	ret = xioctl(video_fd, VIDIOC_QBUF, &buf);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_QBUF");
		goto err;
	}

	ret = tmpret;
 err:
	return ret;
}

static int videodev_stream(struct buffer *mapped_bufs,
			   uint32_t buf_count, uint32_t saves, uint32_t drops,
			   uint8_t timeout_sec, struct image **res_img,
			   uint32_t * ri_size)
{
	int ret;
	uint32_t res_count;
	uint32_t lv;
	struct timeval tv;
	fd_set fds;

	CHECK_VAL(!(mapped_bufs && res_img && ri_size), ret = EINVAL; goto err;);

	drops += 1;		//last one also drops
	res_count = 0;
	*ri_size = 0;

	if (saves) {
		*res_img = (struct image *)malloc(saves * sizeof(**res_img));
		if (!(*res_img)) {
			ret = ENOMEM;
			goto err;
		}
	}

	while (drops) {
		FD_ZERO(&fds);
		FD_SET(video_fd, &fds);

		if (timeout_sec) {
			CLEAR(tv);
			tv.tv_sec = timeout_sec;
			ret = select(video_fd + 1, &fds, NULL, NULL, &tv);
		} else
			ret = select(video_fd + 1, &fds, NULL, NULL, NULL);
		if (-1 == ret) {
			if (EINTR != errno) {
				ret = errno;
				perror("select");
				goto err_free;
			} else if (!ret) {
				fprintf(stderr, "no response from v4l2 core"
					" (select timeout)\n");
				ret = EIO;
				goto err_free;
			}
		} else {
			if ((1 == drops) && saves) {
				ret = videodev_frame_op(mapped_bufs,
							buf_count,
							(*res_img) + res_count);
				if (0 == ret) {
					saves--;
					res_count++;
					(*ri_size)++;
				} else {
					goto err;
				}
			} else {
				ret = videodev_frame_op(mapped_bufs,
							buf_count, NULL);
				if (0 == ret)
					drops--;
			}

			if (0 != ret && ret != EAGAIN) {
				goto err_free_all;
			}
		}
	}

	return 0;
 err_free_all:
	for (lv = 0; lv < res_count; lv++)
		free((*res_img)[lv].data);
 err_free:
	free(*res_img);
 err:
	return ret;
}

static int videodev_capture_stop()
{
	int ret;
	enum v4l2_buf_type bufftype = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	ret = xioctl(video_fd, VIDIOC_STREAMOFF, &bufftype);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_STREAMOFF");
		goto err;
	}

	return 0;
 err:
	return ret;
}

int videodev_std_set(v4l2_std_id std_id)
{
	int ret;
	struct v4l2_input input;

	CLEAR(input);

	ret = xioctl(video_fd, VIDIOC_G_INPUT, &input.index);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_G_INPUT");
		goto done;
	}

	ret = xioctl(video_fd, VIDIOC_ENUMINPUT, &input);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_ENUMINPUT");
		goto done;
	}

	if (0 == (input.std & std_id)) {
		fprintf(stderr, "unsupported video standard\n");
		ret = EPERM;
		goto done;
	}

	ret = xioctl(video_fd, VIDIOC_S_STD, &std_id);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_S_STD");
		goto done;
	}

	ret = 0;
 done:
	return ret;
}

int videodev_capture(uint32_t saves, uint32_t drops,
			 uint8_t timeout_sec, struct image **res_img,
			 uint32_t * ri_size)
{
	int ret;
	struct buffer *buffers;
	uint32_t buf_count;

	CHECK_VAL(!(res_img && ri_size), ret = EINVAL; goto done;);

	buf_count = 0;

	ret = videodev_setup_capture();
	if (ret)
		goto done;

	buffers = videodev_mmap_init(&buf_count);
	if (IS_ERR(buffers)) {
		ret = PTR_ERR(buffers);
		goto done;
	}

	ret = videodev_capture_start(buf_count);
	if (ret)
		goto done_unmap;

	ret =
		videodev_stream(buffers, buf_count, saves, drops, timeout_sec,
				res_img, ri_size);
	if (ret)
		goto done_unmap;

	ret = videodev_capture_stop();
	if (ret)
		free(res_img);

 done_unmap:
	videodev_unmap(buffers, buf_count);
 done:
	if (ret)
		return ret;
	return 0;
}

int videodev_stds_print()
{
	int ret;
	struct v4l2_input inp;
	struct v4l2_standard std;

	CLEAR(inp);
	CLEAR(std);

	ret = xioctl(video_fd, VIDIOC_G_INPUT, &inp.index);
	if (-1 == ret) {
		ret = errno;
		perror("VIDIOC_G_INPUT");
		goto done;
	}

	printf("Current input %s supports:\n", inp.name);

	while (!xioctl(video_fd, VIDIOC_ENUMSTD, &std)) {
		if (std.id & inp.std)
			printf("%s\n", std.name);
		std.index++;
	}

	if (EINVAL != errno || !std.index) {
		ret = errno;
		perror("VIDIOC_ENUMSTD");
		goto done;
	}

	ret = 0;

 done:
	return ret;
}
